@@ -1,19 +1,17 @@
-{-# OPTIONS_GHC -Wall #-}
-
--- | Numbers that can be indexed into with an Int list.
---
---
+-- | Multi-dimensional arrays
 module NumHask.Array
   ( -- * Imports
-    --
     -- $imports
-    --
+
+    -- * Overview
+    -- $overview
     module NumHask.Array.Shape,
     module NumHask.Array.Fixed,
-  ) where
+  )
+where
 
-import NumHask.Array.Shape
 import NumHask.Array.Fixed
+import NumHask.Array.Shape
 
 -- $imports
 --
@@ -26,9 +24,15 @@ import NumHask.Array.Fixed
 -- > import NumHask.Array.Shape
 -- > import qualified NumHask.Array.Fixed as F
 -- > import qualified NumHask.Array.Dynamic as D
+
+-- $overview
 --
--- A hmatrix instance of Array is also provided for performance purposes:
+-- 'Array's are higher-kinded numbers that can be indexed into with an @[Int]@. Higher-kinded numbers are things with a non-primitive type that we wish to use the usual numerical operators on: '+','-','*','/','abs','tan' and so on.
 --
--- > import NumHask.Array.Shape
--- > import qualified NumHask.Array.HMatrix as H
+-- The design of numhask-array:
 --
+-- - allows shape to be specified at both the type and value level.
+--
+-- - provides operators at value and type level to help manipulate shapes.
+--
+-- - provides fixed and dynamic arrays with the same API.
