@@ -1,41 +1,37 @@
 {-# LANGUAGE DataKinds #-}
-{-# LANGUAGE DeriveAnyClass #-}
-{-# LANGUAGE DeriveGeneric #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE NoImplicitPrelude #-}
-{-# LANGUAGE NoStarIsType #-}
-{-# LANGUAGE PolyKinds #-}
-{-# LANGUAGE RankNTypes #-}
-{-# LANGUAGE ScopedTypeVariables #-}
+{-# LANGUAGE RebindableSyntax #-}
 {-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE TypeOperators #-}
-{-# OPTIONS_GHC -Wno-redundant-constraints #-}
+{-# OPTIONS_GHC -Wno-x-partial #-}
 {-# OPTIONS_GHC -fno-warn-incomplete-uni-patterns #-}
 
--- | Arrays with a dynamic shape.
+-- | Arrays with a dynamic shape (shape only known at runtime).
 module NumHask.Array.Dynamic
-  ( -- $setup
+  ( -- $usage
     Array (..),
 
     -- * Conversion
     fromFlatList,
+    toFlatList,
 
     -- * representable replacements
     index,
     tabulate,
 
     -- * Operators
+    takes,
     reshape,
     transpose,
-    diag,
+    indices,
     ident,
+    sequent,
+    diag,
+    undiag,
     singleton,
     selects,
     selectsExcept,
     folds,
     extracts,
+    extractsExcept,
     joins,
     maps,
     concatenate,
@@ -43,6 +39,8 @@ module NumHask.Array.Dynamic
     append,
     reorder,
     expand,
+    expandr,
+    apply,
     contract,
     dot,
     mult,
@@ -50,12 +48,14 @@ module NumHask.Array.Dynamic
     squeeze,
 
     -- * Scalar
+
     --
     -- Scalar specialisations
     fromScalar,
     toScalar,
 
     -- * Matrix
+
     --
     -- Matrix specialisations.
     col,
@@ -64,17 +64,35 @@ module NumHask.Array.Dynamic
   )
 where
 
-import Data.List ((!!))
-import qualified Data.Vector as V
+import Data.List (intercalate)
+import Data.Vector qualified as V
 import GHC.Show (Show (..))
 import NumHask.Array.Shape
-import NumHask.Prelude as P hiding (product, transpose)
+import NumHask.Prelude as P hiding (product)
 
 -- $setup
 -- >>> :set -XDataKinds
 -- >>> :set -XOverloadedLists
 -- >>> :set -XTypeFamilies
 -- >>> :set -XFlexibleContexts
+-- >>> :set -XRebindableSyntax
+-- >>> import NumHask.Prelude
+-- >>> import NumHask.Array.Dynamic
+-- >>> import NumHask.Array.Shape
+-- >>> let s = fromFlatList [] [1] :: Array Int
+-- >>> let a = fromFlatList [2,3,4] [1..24] :: Array Int
+-- >>> let v = fromFlatList [3] [1,2,3] :: Array Int
+-- >>> let m = fromFlatList [3,4] [0..11] :: Array Int
+
+-- $usage
+-- >>> :set -XDataKinds
+-- >>> :set -XOverloadedLists
+-- >>> :set -XTypeFamilies
+-- >>> :set -XFlexibleContexts
+-- >>> :set -XRebindableSyntax
+-- >>> import NumHask.Prelude
+-- >>> import NumHask.Array.Dynamic
+-- >>> import NumHask.Array.Shape
 -- >>> let s = fromFlatList [] [1] :: Array Int
 -- >>> let a = fromFlatList [2,3,4] [1..24] :: Array Int
 -- >>> let v = fromFlatList [3] [1,2,3] :: Array Int
@@ -90,9 +108,8 @@ import NumHask.Prelude as P hiding (product, transpose)
 --  [[13, 14, 15, 16],
 --   [17, 18, 19, 20],
 --   [21, 22, 23, 24]]]
-data Array a
-  = Array {shape :: [Int], unArray :: V.Vector a}
-  deriving (Eq, Ord, NFData, Generic)
+data Array a = Array {shape :: [Int], unArray :: V.Vector a}
+  deriving (Eq, Ord, Generic)
 
 instance Functor Array where
   fmap f (Array s a) = Array s (V.map f a)
@@ -109,7 +126,7 @@ instance (Show a) => Show (Array a) where
     where
       go n a'@(Array l' m) =
         case length l' of
-          0 -> maybe (throw (NumHaskException "empty scalar")) GHC.Show.show (head m)
+          0 -> GHC.Show.show (V.head m)
           1 -> "[" ++ intercalate ", " (GHC.Show.show <$> V.toList m) ++ "]"
           x ->
             "["
@@ -148,6 +165,19 @@ index (Array s v) i = V.unsafeIndex v (flatten s i)
 tabulate :: () => [Int] -> ([Int] -> a) -> Array a
 tabulate ds f = Array ds . V.generate (size ds) $ (f . shapen ds)
 
+-- | Takes the top-most elements according to the new dimension.
+--
+-- >>> takes [2,2,3] a
+-- [[[1, 2, 3],
+--   [5, 6, 7]],
+--  [[13, 14, 15],
+--   [17, 18, 19]]]
+takes ::
+  [Int] ->
+  Array a ->
+  Array a
+takes ds a = tabulate ds $ \s -> index a s
+
 -- | Reshape an array (with the same number of elements).
 --
 -- >>> reshape [4,3,2] a
@@ -176,20 +206,45 @@ reshape s a = tabulate s (index a . shapen (shape a) . flatten s)
 transpose :: Array a -> Array a
 transpose a = tabulate (reverse $ shape a) (index a . reverse)
 
+-- | Indices of an Array.
+--
+-- >>> indices [3,3]
+-- [[[0,0], [0,1], [0,2]],
+--  [[1,0], [1,1], [1,2]],
+--  [[2,0], [2,1], [2,2]]]
+indices :: [Int] -> Array [Int]
+indices ds = tabulate ds id
+
 -- | The identity array.
 --
 -- >>> ident [3,2]
 -- [[1, 0],
 --  [0, 1],
 --  [0, 0]]
-ident :: (Num a) => [Int] -> Array a
-ident ds = tabulate ds (bool 0 1 . isDiag)
+ident :: (Additive a, Multiplicative a) => [Int] -> Array a
+ident ds = tabulate ds (bool zero one . isDiag)
   where
     isDiag [] = True
     isDiag [_] = True
     isDiag [x, y] = x == y
     isDiag (x : y : xs) = x == y && isDiag (y : xs)
 
+-- | An array of sequential Ints
+--
+-- >>> sequent [3]
+-- [0, 1, 2]
+--
+-- >>> sequent [3,3]
+-- [[0, 0, 0],
+--  [0, 1, 0],
+--  [0, 0, 2]]
+sequent :: [Int] -> Array Int
+sequent ds = tabulate ds go
+  where
+    go [] = zero
+    go [i] = i
+    go (i : js) = bool zero i (all (i ==) js)
+
 -- | Extract the diagonal of an array.
 --
 -- >>> diag (ident [3,2])
@@ -202,6 +257,21 @@ diag a = tabulate [NumHask.Array.Shape.minimum (shape a)] go
     go [] = throw (NumHaskException "Rank Underflow")
     go (s' : _) = index a (replicate (rank (shape a)) s')
 
+-- | Expand the array to form a diagonal array
+--
+-- >>> undiag 2 (fromFlatList [2] [1,1])
+-- [[1, 0],
+--  [0, 1]]
+undiag ::
+  (Additive a) =>
+  Int ->
+  Array a ->
+  Array a
+undiag r a = tabulate (replicate r (head (shape a))) go
+  where
+    go [] = throw (NumHaskException "Rank Underflow")
+    go xs@(x : xs') = bool zero (index a xs) (all (x ==) xs')
+
 -- | Create an array composed of a single value.
 --
 -- >>> singleton [3,2] one
@@ -398,6 +468,19 @@ reorder ds a = tabulate (reorder' (shape a) ds) go
 -- [[1, 2, 3],
 --  [2, 4, 6],
 --  [3, 6, 9]]
+--
+-- Alternatively, expand can be understood as representing the permutation of element pairs of two arrays, so like the Applicative List instance.
+--
+-- >>> i2 = indices [2,2]
+-- >>> expand (,) i2 i2
+-- [[[[([0,0],[0,0]), ([0,0],[0,1])],
+--    [([0,0],[1,0]), ([0,0],[1,1])]],
+--   [[([0,1],[0,0]), ([0,1],[0,1])],
+--    [([0,1],[1,0]), ([0,1],[1,1])]]],
+--  [[[([1,0],[0,0]), ([1,0],[0,1])],
+--    [([1,0],[1,0]), ([1,0],[1,1])]],
+--   [[([1,1],[0,0]), ([1,1],[0,1])],
+--    [([1,1],[1,0]), ([1,1],[1,1])]]]]
 expand ::
   (a -> b -> c) ->
   Array a ->
@@ -407,6 +490,51 @@ expand f a b = tabulate ((++) (shape a) (shape b)) (\i -> f (index a (take r i))
   where
     r = rank (shape a)
 
+-- | Like expand, but permutes the first array first, rather than the second.
+--
+-- >>> expand (,) v (fmap (+3) v)
+-- [[(1,4), (1,5), (1,6)],
+--  [(2,4), (2,5), (2,6)],
+--  [(3,4), (3,5), (3,6)]]
+--
+-- >>> expandr (,) v (fmap (+3) v)
+-- [[(1,4), (2,4), (3,4)],
+--  [(1,5), (2,5), (3,5)],
+--  [(1,6), (2,6), (3,6)]]
+expandr ::
+  (a -> b -> c) ->
+  Array a ->
+  Array b ->
+  Array c
+expandr f a b = tabulate ((++) (shape a) (shape b)) (\i -> f (index a (drop r i)) (index b (take r i)))
+  where
+    r = rank (shape a)
+
+-- | Apply an array of functions to each array of values.
+--
+-- This is in the spirit of the applicative functor operation (\<*\>).
+--
+-- > expand f a b == apply (fmap f a) b
+--
+-- >>> apply ((*) <$> v) v
+-- [[1, 2, 3],
+--  [2, 4, 6],
+--  [3, 6, 9]]
+--
+-- Dynamic arrays can't be Applicatives because there is no 'pure' (Shape is not known at compile-time).
+--
+-- >>> let b = fromFlatList [2,3] [1..6] :: Array Int
+-- >>> contract sum [1,2] (apply (fmap (*) b) (transpose b))
+-- [[14, 32],
+--  [32, 77]]
+apply ::
+  Array (a -> b) ->
+  Array a ->
+  Array b
+apply f a = tabulate ((++) (shape f) (shape a)) (\i -> index f (take r i) (index a (drop r i)))
+  where
+    r = rank (shape f)
+
 -- | Contract an array by applying the supplied (folding) function on diagonal elements of the dimensions.
 --
 -- This generalises a tensor contraction by allowing the number of contracting diagonals to be other than 2, and allowing a binary operator other than multiplication.
@@ -523,7 +651,7 @@ fromScalar a = index a ([] :: [Int])
 -- | Convert a number to a scalar.
 --
 -- >>> :t toScalar 2
--- toScalar 2 :: Num a => Array a
+-- toScalar 2 :: FromInteger a => Array a
 toScalar :: a -> Array a
 toScalar a = fromFlatList [] [a]
 
