@@ -1,24 +1,12 @@
 {-# LANGUAGE DataKinds #-}
-{-# LANGUAGE DeriveGeneric #-}
-{-# LANGUAGE DeriveTraversable #-}
-{-# LANGUAGE FlexibleContexts #-}
-{-# LANGUAGE FlexibleInstances #-}
-{-# LANGUAGE GeneralizedNewtypeDeriving #-}
-{-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE NoImplicitPrelude #-}
-{-# LANGUAGE NoStarIsType #-}
-{-# LANGUAGE PolyKinds #-}
-{-# LANGUAGE RankNTypes #-}
-{-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE TypeApplications #-}
+{-# LANGUAGE RebindableSyntax #-}
 {-# LANGUAGE TypeFamilies #-}
-{-# LANGUAGE TypeOperators #-}
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
 {-# OPTIONS_GHC -Wno-redundant-constraints #-}
-{-# OPTIONS_GHC -fno-warn-unused-imports #-}
 
--- | Arrays with a fixed shape.
+-- | Arrays with a fixed shape (known shape at compile time).
 module NumHask.Array.Fixed
-  ( -- $setup
+  ( -- $usage
     Array (..),
 
     -- * Conversion
@@ -27,15 +15,20 @@ module NumHask.Array.Fixed
     toDynamic,
 
     -- * Operators
+    takes,
     reshape,
     transpose,
-    diag,
+    indices,
     ident,
+    sequent,
+    diag,
+    undiag,
     singleton,
     selects,
     selectsExcept,
     folds,
     extracts,
+    extractsExcept,
     joins,
     maps,
     concatenate,
@@ -43,6 +36,8 @@ module NumHask.Array.Fixed
     append,
     reorder,
     expand,
+    expandr,
+    apply,
     contract,
     dot,
     mult,
@@ -50,18 +45,21 @@ module NumHask.Array.Fixed
     squeeze,
 
     -- * Scalar
+
     --
     -- Scalar specialisations
-    Scalar,
     fromScalar,
     toScalar,
 
     -- * Vector
+
     --
     -- Vector specialisations.
     Vector,
+    sequentv,
 
     -- * Matrix
+
     --
     -- Matrix specialisations.
     Matrix,
@@ -70,26 +68,49 @@ module NumHask.Array.Fixed
     safeCol,
     safeRow,
     mmult,
+    chol,
+    invtri,
   )
 where
 
-import Control.Category (id)
 import Data.Distributive (Distributive (..))
 import Data.Functor.Rep
-import Data.List ((!!))
-import qualified Data.Vector as V
+import Data.Proxy
+import Data.Vector qualified as V
 import GHC.Exts (IsList (..))
 import GHC.Show (Show (..))
 import GHC.TypeLits
-import qualified NumHask.Array.Dynamic as D
+import NumHask.Array.Dynamic qualified as D
 import NumHask.Array.Shape
-import NumHask.Prelude as P hiding (identity, transpose)
+import NumHask.Prelude as P hiding (sequence, toList)
 
 -- $setup
+--
 -- >>> :set -XDataKinds
 -- >>> :set -XOverloadedLists
 -- >>> :set -XTypeFamilies
 -- >>> :set -XFlexibleContexts
+-- >>> :set -XRebindableSyntax
+-- >>> import NumHask.Prelude
+-- >>> import GHC.TypeLits (Nat)
+-- >>> import Data.Proxy
+-- >>> import Data.Functor.Rep
+-- >>> let s = [1] :: Array ('[] :: [Nat]) Int -- scalar
+-- >>> let v = [1,2,3] :: Array '[3] Int       -- vector
+-- >>> let t = [0..3] :: Array '[2,2] Int     -- square matrix
+-- >>> let m = [0..11] :: Array '[3,4] Int     -- matrix
+-- >>> let a = [1..24] :: Array '[2,3,4] Int
+
+-- $usage
+--
+-- >>> :set -XDataKinds
+-- >>> :set -XOverloadedLists
+-- >>> :set -XTypeFamilies
+-- >>> :set -XFlexibleContexts
+-- >>> :set -XRebindableSyntax
+-- >>> import NumHask.Prelude
+-- >>> import NumHask.Array.Fixed
+-- >>> import GHC.TypeLits (Nat)
 -- >>> let s = [1] :: Array ('[] :: [Nat]) Int -- scalar
 -- >>> let v = [1,2,3] :: Array '[3] Int       -- vector
 -- >>> let m = [0..11] :: Array '[3,4] Int     -- matrix
@@ -97,8 +118,8 @@ import NumHask.Prelude as P hiding (identity, transpose)
 
 -- | a multidimensional array with a type-level shape
 --
--- >>> let a = [1..24] :: Array '[2,3,4] Int
--- >>> a
+-- >>> :set -XDataKinds
+-- >>> [1..24] :: Array '[2,3,4] Int
 -- [[[1, 2, 3, 4],
 --   [5, 6, 7, 8],
 --   [9, 10, 11, 12]],
@@ -107,8 +128,9 @@ import NumHask.Prelude as P hiding (identity, transpose)
 --   [21, 22, 23, 24]]]
 --
 -- >>> [1,2,3] :: Array '[2,2] Int
--- [[*** Exception: NumHaskException {errorMessage = "shape mismatch"}
-newtype Array s a = Array {unArray :: V.Vector a} deriving (Eq, Ord, NFData, Functor, Foldable, Generic, Traversable)
+-- *** Exception: NumHaskException {errorMessage = "shape mismatch"}
+-- [[
+newtype Array s a = Array {unArray :: V.Vector a} deriving (Eq, Ord, Functor, Foldable, Generic, Traversable)
 
 instance (HasShape s, Show a) => Show (Array s a) where
   show a = GHC.Show.show (toDynamic a)
@@ -127,7 +149,6 @@ instance
   ) =>
   Representable (Array s)
   where
-
   type Rep (Array s) = [Int]
 
   tabulate f =
@@ -149,7 +170,6 @@ instance
   ) =>
   Additive (Array s a)
   where
-
   (+) = liftR2 (+)
 
   zero = pureRep zero
@@ -162,34 +182,28 @@ instance
   where
   negate = fmapRep negate
 
-type instance Actor (Array s a) = a
-
 instance
-  ( Multiplicative a,
-    HasShape s
-  ) =>
-  HadamardMultiplication (Array s) a
-  where
-  (.*.) = liftR2 (*)
-
-instance
-  ( Divisive a,
-    HasShape s
-  ) =>
-  HadamardDivision (Array s) a
-  where
-  (./.) = liftR2 (/)
-
-instance
-  (HasShape s, Multiplicative a) =>
+  (Multiplicative a) =>
   MultiplicativeAction (Array s a)
   where
+  type Scalar (Array s a) = a
+  (|*) r s = fmap (s *) r
 
-  (.*) r s = fmap (* s) r
-  {-# INLINE (.*) #-}
+instance (Additive a) => AdditiveAction (Array s a) where
+  type AdditiveScalar (Array s a) = a
+  (|+) r s = fmap (s +) r
 
-  (*.) s = fmap (s *)
-  {-# INLINE (*.) #-}
+instance
+  (Subtractive a) =>
+  SubtractiveAction (Array s a)
+  where
+  (|-) r s = fmap (\x -> x - s) r
+
+instance
+  (Divisive a) =>
+  DivisiveAction (Array s a)
+  where
+  (|/) r s = fmap (/ s) r
 
 instance (HasShape s, JoinSemiLattice a) => JoinSemiLattice (Array s a) where
   (\/) = liftR2 (\/)
@@ -198,17 +212,13 @@ instance (HasShape s, MeetSemiLattice a) => MeetSemiLattice (Array s a) where
   (/\) = liftR2 (/\)
 
 instance (HasShape s, Subtractive a, Epsilon a) => Epsilon (Array s a) where
-
   epsilon = singleton epsilon
 
-  nearZero (Array a) = all nearZero a
-
 instance
   ( HasShape s
   ) =>
   IsList (Array s a)
   where
-
   type Item (Array s a) = a
 
   fromList l =
@@ -231,10 +241,11 @@ shape _ = shapeVal $ toShape @s
 
 -- | convert to a dynamic array with shape at the value level.
 toDynamic :: (HasShape s) => Array s a -> D.Array a
-toDynamic a = D.fromFlatList (shape a) (P.toList a)
+toDynamic a = D.fromFlatList (shape a) (toList a)
 
 -- | Use a dynamic array in a fixed context.
 --
+-- >>> import qualified NumHask.Array.Dynamic as D
 -- >>> with (D.fromFlatList [2,3,4] [1..24]) (selects (Proxy :: Proxy '[0,1]) [1,1] :: Array '[2,3,4] Int -> Array '[4] Int)
 -- [17, 18, 19, 20]
 with ::
@@ -245,6 +256,22 @@ with ::
   r
 with (D.Array _ v) f = f (Array v)
 
+-- | Takes the top-most elements according to the new dimension.
+--
+-- >>> takes a :: Array '[2,2,3] Int
+-- [[[1, 2, 3],
+--   [5, 6, 7]],
+--  [[13, 14, 15],
+--   [17, 18, 19]]]
+takes ::
+  forall s s' a.
+  ( HasShape s,
+    HasShape s'
+  ) =>
+  Array s a ->
+  Array s' a
+takes a = tabulate $ \s -> index a s
+
 -- | Reshape an array (with the same number of elements).
 --
 -- >>> reshape a :: Array '[4,3,2] Int
@@ -280,6 +307,15 @@ reshape a = tabulate (index a . shapen s . flatten s')
 transpose :: forall a s. (HasShape s, HasShape (Reverse s)) => Array s a -> Array (Reverse s) a
 transpose a = tabulate (index a . reverse)
 
+-- | Indices of an Array.
+--
+-- >>> indices :: Array '[3,3] [Int]
+-- [[[0,0], [0,1], [0,2]],
+--  [[1,0], [1,1], [1,2]],
+--  [[2,0], [2,1], [2,2]]]
+indices :: forall s. (HasShape s) => Array s [Int]
+indices = tabulate id
+
 -- | The identity array.
 --
 -- >>> ident :: Array '[3,2] Int
@@ -294,6 +330,22 @@ ident = tabulate (bool zero one . isDiag)
     isDiag [x, y] = x == y
     isDiag (x : y : xs) = x == y && isDiag (y : xs)
 
+-- | An array of sequential Ints
+--
+-- >>> sequent :: Array '[3] Int
+-- [0, 1, 2]
+--
+-- >>> sequent :: Array '[3,3] Int
+-- [[0, 0, 0],
+--  [0, 1, 0],
+--  [0, 0, 2]]
+sequent :: forall s. (HasShape s) => Array s Int
+sequent = tabulate go
+  where
+    go [] = zero
+    go [i] = i
+    go (i : js) = bool zero i (all (i ==) js)
+
 -- | Extract the diagonal of an array.
 --
 -- >>> diag (ident :: Array '[3,2] Int)
@@ -311,6 +363,24 @@ diag a = tabulate go
     go (s' : _) = index a (replicate (length ds) s')
     ds = shapeVal (toShape @s)
 
+-- | Expand the array to form a diagonal array
+--
+-- >>> undiag ([1,1] :: Array '[2] Int)
+-- [[1, 0],
+--  [0, 1]]
+undiag ::
+  forall a s.
+  ( HasShape s,
+    Additive a,
+    HasShape ((++) s s)
+  ) =>
+  Array s a ->
+  Array ((++) s s) a
+undiag a = tabulate go
+  where
+    go [] = throw (NumHaskException "Rank Underflow")
+    go xs@(x : xs') = bool zero (index a xs) (all (x ==) xs')
+
 -- | Create an array composed of a single value.
 --
 -- >>> singleton one :: Array '[3,2] Int
@@ -393,7 +463,7 @@ folds f d a = tabulate go
 --
 -- >>> let e = extracts (Proxy :: Proxy '[1,2]) a
 -- >>> :t e
--- e :: Array '[3, 4] (Array '[2] Int)
+-- e :: Array [3, 4] (Array '[2] Int)
 extracts ::
   forall ds st si so a.
   ( HasShape st,
@@ -414,7 +484,7 @@ extracts d a = tabulate go
 --
 -- >>> let e = extractsExcept (Proxy :: Proxy '[1,2]) a
 -- >>> :t e
--- e :: Array '[2] (Array '[3, 4] Int)
+-- e :: Array '[2] (Array [3, 4] Int)
 extractsExcept ::
   forall ds st si so a.
   ( HasShape st,
@@ -436,12 +506,12 @@ extractsExcept d a = tabulate go
 -- >>> let e = extracts (Proxy :: Proxy '[1,0]) a
 --
 -- >>> :t e
--- e :: Array '[3, 2] (Array '[4] Int)
+-- e :: Array [3, 2] (Array '[4] Int)
 --
 -- >>> let j = joins (Proxy :: Proxy '[1,0]) e
 --
 -- >>> :t j
--- j :: Array '[2, 3, 4] Int
+-- j :: Array [2, 3, 4] Int
 --
 -- >>> a == j
 -- True
@@ -464,7 +534,7 @@ joins _ a = tabulate go
 -- | Maps a function along specified dimensions.
 --
 -- >>> :t maps (transpose) (Proxy :: Proxy '[1]) a
--- maps (transpose) (Proxy :: Proxy '[1]) a :: Array '[4, 3, 2] Int
+-- maps (transpose) (Proxy :: Proxy '[1]) a :: Array [4, 3, 2] Int
 maps ::
   forall ds st st' si si' so a b.
   ( HasShape st,
@@ -487,7 +557,7 @@ maps f d a = joins d (fmapRep f (extracts d a))
 -- | Concatenate along a dimension.
 --
 -- >>> :t concatenate (Proxy :: Proxy 1) a a
--- concatenate (Proxy :: Proxy 1) a a :: Array '[2, 6, 4] Int
+-- concatenate (Proxy :: Proxy 1) a a :: Array [2, 6, 4] Int
 concatenate ::
   forall a s0 s1 d s.
   ( CheckConcatenate d s0 s1 s,
@@ -555,7 +625,7 @@ insert _ _ a b = tabulate go
 --
 -- >>>  :t append (Proxy :: Proxy 0) a
 -- append (Proxy :: Proxy 0) a
---   :: Array '[3, 4] Int -> Array '[3, 3, 4] Int
+--   :: Array [3, 4] Int -> Array [3, 3, 4] Int
 append ::
   forall a d s s'.
   ( DropIndex s d ~ s',
@@ -576,7 +646,7 @@ append d = insert d (Proxy :: Proxy (Dimension s d - 1))
 --
 -- >>> let r = reorder (Proxy :: Proxy '[2,0,1]) a
 -- >>> :t r
--- r :: Array '[4, 2, 3] Int
+-- r :: Array [4, 2, 3] Int
 reorder ::
   forall a ds s.
   ( HasShape ds,
@@ -597,7 +667,7 @@ reorder _ a = tabulate go
 -- For context, if the function is multiply, and the arrays are tensors,
 -- then this can be interpreted as a tensor product.
 --
--- https://en.wikipedia.org/wiki/Tensor_product
+-- < https://en.wikipedia.org/wiki/Tensor_product>
 --
 -- The concept of a tensor product is a dense crossroad, and a complete treatment is elsewhere.  To quote:
 --
@@ -607,6 +677,19 @@ reorder _ a = tabulate go
 -- [[1, 2, 3],
 --  [2, 4, 6],
 --  [3, 6, 9]]
+--
+-- Alternatively, expand can be understood as representing the permutation of element pairs of two arrays, so like the Applicative List instance.
+--
+-- >>> i2 = indices :: Array '[2,2] [Int]
+-- >>> expand (,) i2 i2
+-- [[[[([0,0],[0,0]), ([0,0],[0,1])],
+--    [([0,0],[1,0]), ([0,0],[1,1])]],
+--   [[([0,1],[0,0]), ([0,1],[0,1])],
+--    [([0,1],[1,0]), ([0,1],[1,1])]]],
+--  [[[([1,0],[0,0]), ([1,0],[0,1])],
+--    [([1,0],[1,0]), ([1,0],[1,1])]],
+--   [[([1,1],[0,0]), ([1,1],[0,1])],
+--    [([1,1],[1,0]), ([1,1],[1,1])]]]]
 expand ::
   forall s s' a b c.
   ( HasShape s,
@@ -621,6 +704,68 @@ expand f a b = tabulate (\i -> f (index a (take r i)) (index b (drop r i)))
   where
     r = rank (shape a)
 
+-- | Like expand, but permutes the first array first, rather than the second.
+--
+-- >>> expand (,) v (v |+ 3)
+-- [[(1,4), (1,5), (1,6)],
+--  [(2,4), (2,5), (2,6)],
+--  [(3,4), (3,5), (3,6)]]
+--
+-- >>> expandr (,) v (v |+ 3)
+-- [[(1,4), (2,4), (3,4)],
+--  [(1,5), (2,5), (3,5)],
+--  [(1,6), (2,6), (3,6)]]
+expandr ::
+  forall s s' a b c.
+  ( HasShape s,
+    HasShape s',
+    HasShape ((++) s s')
+  ) =>
+  (a -> b -> c) ->
+  Array s a ->
+  Array s' b ->
+  Array ((++) s s') c
+expandr f a b = tabulate (\i -> f (index a (drop r i)) (index b (take r i)))
+  where
+    r = rank (shape a)
+
+-- | Apply an array of functions to each array of values.
+--
+-- This is in the spirit of the applicative functor operation (\<*\>).
+--
+-- > expand f a b == apply (fmap f a) b
+--
+-- >>> apply ((*) <$> v) v
+-- [[1, 2, 3],
+--  [2, 4, 6],
+--  [3, 6, 9]]
+--
+-- Fixed Arrays can't be applicative functors because the changes in shape are reflected in the types.
+--
+-- > :t apply
+-- > apply
+-- >   :: (HasShape s, HasShape s', HasShape (s ++ s')) =>
+-- >      Array s (a -> b) -> Array s' a -> Array (s ++ s') b
+-- > :t (<*>)
+-- > (<*>) :: Applicative f => f (a -> b) -> f a -> f b
+--
+-- >>> let b = [1..6] :: Array '[2,3] Int
+-- >>> contract sum (Proxy :: Proxy '[1,2]) (apply (fmap (*) b) (transpose b))
+-- [[14, 32],
+--  [32, 77]]
+apply ::
+  forall s s' a b.
+  ( HasShape s,
+    HasShape s',
+    HasShape ((++) s s')
+  ) =>
+  Array s (a -> b) ->
+  Array s' a ->
+  Array ((++) s s') b
+apply f a = tabulate (\i -> index f (take r i) (index a (drop r i)))
+  where
+    r = rank (shape f)
+
 -- | Contract an array by applying the supplied (folding) function on diagonal elements of the dimensions.
 --
 -- This generalises a tensor contraction by allowing the number of contracting diagonals to be other than 2, and allowing a binary operator other than multiplication.
@@ -673,24 +818,31 @@ contract f xs a = f . diag <$> extractsExcept xs a
 -- >>> dot sum (*) b v
 -- [14, 32]
 --
--- dot allows operation on mis-shaped matrices:
+-- Array elements don't have to be numbers:
+--
+-- >>> x1 = (show <$> [1..4]) :: Array '[2,2] String
+-- >>> x2 = (show <$> [5..8]) :: Array '[2,2] String
+-- >>> x1
+-- [["1", "2"],
+--  ["3", "4"]]
+--
+-- >>> x2
+-- [["5", "6"],
+--  ["7", "8"]]
+--
+-- >>> import Data.List (intercalate)
+-- >>> dot (intercalate "+" . toList) (\a b -> a <> "*" <> b) x1 x2
+-- [["1*5+2*7", "1*6+2*8"],
+--  ["3*5+4*7", "3*6+4*8"]]
+--
+-- 'dot' allows operation on mis-shaped matrices. The algorithm ignores excess positions within the contracting dimension(s):
 --
 -- >>> let m23 = [1..6] :: Array '[2,3] Int
 -- >>> let m12 = [1,2] :: Array '[1,2] Int
 -- >>> shape $ dot sum (*) m23 m12
 -- [2,2]
 --
--- the algorithm ignores excess positions within the contracting dimension(s):
---
--- m23 shape: 2 3
---
--- m12 shape: 1 2
---
--- res shape: 2 2
---
--- FIXME: work out whether this is a feature or a bug...
---
--- find instances of a vector in a matrix
+-- Find instances of a vector in a matrix
 --
 -- >>> let cs = fromList ("abacbaab" :: [Char]) :: Array '[4,2] Char
 -- >>> let v = fromList ("ab" :: [Char]) :: Vector 2 Char
@@ -769,7 +921,7 @@ mult = dot sum (*)
 --
 -- >>> let s = slice (Proxy :: Proxy '[[0,1],[0,2],[1,2]]) a
 -- >>> :t s
--- s :: Array '[2, 2, 2] Int
+-- s :: Array [2, 2, 2] Int
 --
 -- >>> s
 -- [[[2, 3],
@@ -847,11 +999,6 @@ squeeze (Array x) = Array x
 -- $scalar
 -- Scalar specialisations
 
--- | <https://en.wikipedia.org/wiki/Scalarr_(mathematics) Wiki Scalar>
---
--- An Array '[] a despite being a Scalar is never-the-less a one-element vector under the hood. Unification of representation is unexplored.
-type Scalar a = Array ('[] :: [Nat]) a
-
 -- | Unwrapping scalars is probably a performance bottleneck.
 --
 -- >>> let s = [3] :: Array ('[] :: [Nat]) Int
@@ -863,13 +1010,17 @@ fromScalar a = index a ([] :: [Int])
 -- | Convert a number to a scalar.
 --
 -- >>> :t toScalar 2
--- toScalar 2 :: Num a => Array '[] a
+-- toScalar 2 :: FromInteger a => Array '[] a
 toScalar :: (HasShape ('[] :: [Nat])) => a -> Array ('[] :: [Nat]) a
 toScalar a = fromList [a]
 
 -- | <https://en.wikipedia.org/wiki/Vector_(mathematics_and_physics) Wiki Vector>
 type Vector s a = Array '[s] a
 
+-- | Vector specialisation of 'sequent'
+sequentv :: forall n. (KnownNat n) => Vector n Int
+sequentv = sequent
+
 -- | <https://en.wikipedia.org/wiki/Matrix_(mathematics) Wiki Matrix>
 type Matrix m n a = Array '[m, n] a
 
@@ -882,11 +1033,61 @@ instance
   ) =>
   Multiplicative (Matrix m m a)
   where
-
   (*) = mmult
 
   one = ident
 
+instance
+  ( Multiplicative a,
+    P.Distributive a,
+    Subtractive a,
+    Eq a,
+    ExpField a,
+    KnownNat m,
+    HasShape '[m, m]
+  ) =>
+  Divisive (Matrix m m a)
+  where
+  recip a = invtri (transpose (chol a)) * invtri (chol a)
+
+-- | <https://math.stackexchange.com/questions/1003801/inverse-of-an-invertible-upper-triangular-matrix-of-order-3 Inverse of a triangular> matrix.
+invtri :: forall a n. (KnownNat n, ExpField a, Eq a) => Array '[n, n] a -> Array '[n, n] a
+invtri a = sum (fmap (l ^) (sequentv :: Vector n Int)) * ti
+  where
+    ti = undiag (fmap recip (diag a))
+    tl = a - undiag (diag a)
+    l = negate (ti * tl)
+
+-- | cholesky decomposition
+--
+-- Uses the <https://en.wikipedia.org/wiki/Cholesky_decomposition#The_Cholesky_algorithm Cholesky-Crout> algorithm.
+chol :: (KnownNat n, ExpField a) => Array '[n, n] a -> Array '[n, n] a
+chol a =
+  let l =
+        tabulate
+          ( \[i, j] ->
+              bool
+                ( one
+                    / index l [j, j]
+                    * ( index a [i, j]
+                          - sum
+                            ( (\k -> index l [i, k] * index l [j, k])
+                                <$> ([zero .. (j - one)] :: [Int])
+                            )
+                      )
+                )
+                ( sqrt
+                    ( index a [i, i]
+                        - sum
+                          ( (\k -> index l [j, k] ^ 2)
+                              <$> ([zero .. (j - one)] :: [Int])
+                          )
+                    )
+                )
+                (i == j)
+          )
+   in l
+
 -- | Extract specialised to a matrix.
 --
 -- >>> row 1 m
